@@ -1,9 +1,12 @@
 from ._array import Array as Array
 from ._array import ArrayEvent as ArrayEvent
+from ._array import TypedArray as TypedArray
 from ._awareness import Awareness as Awareness
 from ._doc import Doc as Doc
+from ._doc import TypedDoc as TypedDoc
 from ._map import Map as Map
 from ._map import MapEvent as MapEvent
+from ._map import TypedMap as TypedMap
 from ._pycrdt import StackItem as StackItem
 from ._pycrdt import SubdocsEvent as SubdocsEvent
 from ._pycrdt import Subscription as Subscription
@@ -28,3 +31,8 @@ from ._undo import UndoManager as UndoManager
 from ._update import get_state as get_state
 from ._update import get_update as get_update
 from ._update import merge_updates as merge_updates
+from ._version import __version__ as __version__
+from ._xml import XmlElement as XmlElement
+from ._xml import XmlEvent as XmlEvent
+from ._xml import XmlFragment as XmlFragment
+from ._xml import XmlText as XmlText
