@@ -1,17 +1,21 @@
 from __future__ import annotations
 
-from typing import Any, Callable, Iterable, Type, TypeVar, cast
+from functools import partial
+from typing import Any, Callable, Generic, Iterable, Literal, Type, TypeVar, Union, cast, overload
 
-from ._base import BaseDoc, BaseType, base_types
+from anyio import BrokenResourceError, create_memory_object_stream
+from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
+
+from ._base import BaseDoc, BaseType, Typed, base_types, forbid_read_transaction
 from ._pycrdt import Doc as _Doc
 from ._pycrdt import SubdocsEvent, Subscription, TransactionEvent
 from ._pycrdt import Transaction as _Transaction
 from ._transaction import NewTransaction, ReadTransaction, Transaction
 
-T_BaseType = TypeVar("T_BaseType", bound=BaseType)
+T = TypeVar("T", bound=BaseType)
 
 
-class Doc(BaseDoc):
+class Doc(BaseDoc, Generic[T]):
     """
     A shared document.
 
@@ -23,7 +27,7 @@ class Doc(BaseDoc):
 
     def __init__(
         self,
-        init: dict[str, BaseType] = {},
+        init: dict[str, T] = {},
         *,
         client_id: int | None = None,
         doc: _Doc | None = None,
@@ -43,6 +47,10 @@ class Doc(BaseDoc):
             self[k] = v
         if Model is not None:
             self._twin_doc = Doc(init)
+        self._send_streams: dict[
+            bool, set[MemoryObjectSendStream[TransactionEvent | SubdocsEvent]]
+        ] = {False: set(), True: set()}
+        self._event_subscription: dict[bool, Subscription] = {}
 
     @property
     def guid(self) -> int:
@@ -160,9 +168,12 @@ class Doc(BaseDoc):
             except Exception as e:
                 self._twin_doc = Doc(dict(self))
                 raise e
-        self._doc.apply_update(update)
+        with self.transaction() as txn:
+            forbid_read_transaction(txn)
+            assert txn._txn is not None
+            self._doc.apply_update(txn._txn, update)
 
-    def __setitem__(self, key: str, value: BaseType) -> None:
+    def __setitem__(self, key: str, value: T) -> None:
         """
         Sets a document root type:
         ```py
@@ -178,11 +189,13 @@ class Doc(BaseDoc):
         """
         if not isinstance(key, str):
             raise RuntimeError("Key must be of type string")
-        integrated = value._get_or_insert(key, self)
-        prelim = value._integrate(self, integrated)
-        value._init(prelim)
+        with self.transaction() as txn:
+            forbid_read_transaction(txn)
+            integrated = value._get_or_insert(key, self)
+            prelim = value._integrate(self, integrated)
+            value._init(prelim)
 
-    def __getitem__(self, key: str) -> BaseType:
+    def __getitem__(self, key: str) -> T:
         """
         Gets the document root type corresponding to the given key:
         ```py
@@ -204,7 +217,7 @@ class Doc(BaseDoc):
         """
         return iter(self.keys())
 
-    def get(self, key: str, *, type: type[T_BaseType]) -> T_BaseType:
+    def get(self, key: str, *, type: type[T]) -> T:
         """
         Gets the document root type corresponding to the given key.
         If it already exists, it will be cast to the given type (if different),
@@ -227,14 +240,14 @@ class Doc(BaseDoc):
         """
         return self._roots.keys()
 
-    def values(self) -> Iterable[BaseType]:
+    def values(self) -> Iterable[T]:
         """
         Returns:
             An iterable over the document root types.
         """
         return self._roots.values()
 
-    def items(self) -> Iterable[tuple[str, BaseType]]:
+    def items(self) -> Iterable[tuple[str, T]]:
         """
         Returns:
             An iterable over the key-value pairs of document root types.
@@ -242,14 +255,14 @@ class Doc(BaseDoc):
         return self._roots.items()
 
     @property
-    def _roots(self) -> dict[str, BaseType]:
+    def _roots(self) -> dict[str, T]:
         with self.transaction() as txn:
             assert txn._txn is not None
             return {
                 key: (
                     None
                     if val is None
-                    else cast(Type[BaseType], base_types[type(val)])(_integrated=val, _doc=self)
+                    else cast(Type[T], base_types[type(val)])(_integrated=val, _doc=self)
                 )
                 for key, val in self._doc.roots(txn._txn).items()
             }
@@ -292,5 +305,108 @@ class Doc(BaseDoc):
         self._subscriptions.remove(subscription)
         subscription.drop()
 
+    @overload
+    def events(
+        self,
+        subdocs: Literal[False],
+        max_buffer_size: float = float("inf"),
+    ) -> MemoryObjectReceiveStream[TransactionEvent]: ...
+
+    @overload
+    def events(
+        self,
+        subdocs: Literal[True],
+        max_buffer_size: float = float("inf"),
+    ) -> MemoryObjectReceiveStream[list[SubdocsEvent]]: ...
+
+    def events(
+        self,
+        subdocs: bool = False,
+        max_buffer_size: float = float("inf"),
+    ):
+        """
+        Allows to asynchronously iterate over the document events, without using a callback.
+        A buffer is used to store the events, allowing to iterate at a (temporarily) slower
+        rate than they are produced.
+
+        This method must be used with an async context manager and an async for-loop:
+
+        ```py
+        async def main():
+            async with doc.events() as events:
+                async for event in events:
+                    update: bytes = event.update
+                    ...
+        ```
+
+        Args:
+            subdocs: Whether to iterate over the [SubdocsEvent][pycrdt.SubdocsEvent] events
+                (default is [TransactionEvent][pycrdt.TransactionEvent]).
+            max_buffer_size: Maximum number of events that can be buffered.
+
+        Returns:
+            An async iterator over the document events.
+        """
+        observe = self.observe_subdocs if subdocs else self.observe
+        if not self._send_streams[subdocs]:
+            self._event_subscription[subdocs] = observe(partial(self._send_event, subdocs))
+        send_stream, receive_stream = create_memory_object_stream[
+            Union[TransactionEvent, SubdocsEvent]
+        ](max_buffer_size=max_buffer_size)
+        self._send_streams[subdocs].add(send_stream)
+        return receive_stream
+
+    def _send_event(self, subdocs: bool, event: TransactionEvent | SubdocsEvent):
+        to_remove: list[MemoryObjectSendStream[TransactionEvent | SubdocsEvent]] = []
+        send_streams = self._send_streams[subdocs]
+        for send_stream in send_streams:
+            try:
+                send_stream.send_nowait(event)
+            except BrokenResourceError:
+                to_remove.append(send_stream)
+        for send_stream in to_remove:
+            send_streams.remove(send_stream)
+        if not send_streams:
+            self.unobserve(self._event_subscription[subdocs])
+
+
+class TypedDoc(Typed):
+    """
+    A container for a [Doc][pycrdt.Doc.__init__] where root shared values have types associated
+    with specific keys. The underlying `Doc` can be accessed with the special `_` attribute.
+
+    ```py
+    from pycrdt import Array, Doc, Map, Text, TypedDoc
+
+    class MyDoc(TypedDoc):
+        map0: Map[int]
+        array0: Array[bool]
+        text0: Text
+
+    doc = MyDoc()
+
+    doc.map0["foo"] = 3
+    doc.array0.append(True)
+    doc.text0 += "Hello"
+    untyped_doc: Doc = doc._
+    ```
+    """
+
+    _: Doc
+
+    def __init__(self, doc: TypedDoc | Doc | None = None) -> None:
+        super().__init__()
+        if doc is None:
+            doc = Doc()
+        elif isinstance(doc, TypedDoc):
+            doc = doc._
+        assert isinstance(doc, Doc)
+        self._ = doc
+        for name, _type in self.__dict__["annotations"].items():
+            root_type = _type()
+            if isinstance(root_type, Typed):
+                root_type = root_type._
+            doc[name] = root_type
+
 
 base_types[_Doc] = Doc
